#!/usr/bin/python
#demography.py

'''
Functions to implement demographic operations (birth and death).
'''

#geonomics imports
from geonomics.ops.selection import _calc_prob_death
from geonomics.utils.viz import _check_display

#other imports
import numpy as np
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from numpy import random as r
from scipy import interpolate
from sklearn.preprocessing import normalize
from collections import OrderedDict as OD
from operator import itemgetter


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

class _DebugPlotter:
    def __init__(self, land, spp, timestep):
        self.subplot = 241
        self.timestep = timestep
        self.spp = spp
        self.land = land

    def _next_plot(self, varname, var):
        if self.subplot <= 248:
            if not plt.get_fignums():
                fig = plt.figure()
                plt.suptitle('TIMESTEP: %i' % self.timestep)
            else:
                fig = plt.gcf()
            plt.subplot(self.subplot)
            plt.title(varname)
            self.spp._plot(land = self.land, lyr_num = 0, size = 8, hide_land = True)
            plt.imshow(var, interpolation = 'nearest', cmap = 'terrain')
            plt.colorbar()
            self.subplot += 1
        else:
            pass


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

def _calc_n_pairs(pairs, spp):
    #if there are pairs
    if pairs.size > 0:
        #get their coordinates
        pairs_coords = spp._get_coords(individs = pairs.flatten())
        #take just the x and y coordinates from pairs_coords, reshape them to
        #match pairs.shape (i.e.nx2, where n = len(spp)), and then mean across
        #axis 1 (to get the pair's midpoint
        #x and y coordinates)
        p_x = pairs_coords[:,0].reshape(pairs.shape).mean(axis = 1)
        p_y = pairs_coords[:,1].reshape(pairs.shape).mean(axis = 1)

        #use the spp._dens_grids to calculate a raster of the expected number
        #of pairs
        #NOTE: because the window_width of that object is much wider than 
        #dispersal_mu in the parameterizations I've typically been running, 
        #this produces much wider areas of high values in the N_b, N_d, and
        #hence d rasters than the old method did (which had its window_width
        #set to dispersal_mu, leading of course to huge lags in the old 
        #calc_density algorithm) need to think about whether I need to attend
        #to this, and if so, how
        n_pairs = np.clip(spp._dens_grids._calc_density(p_x, p_y), a_min = 0,
                                                                a_max = None)

        #remove NaNs
        n_pairs[np.isnan(n_pairs)] = 0

    #otherwise just return a 0 raster
    else:
        n_pairs = np.zeros(spp._land_dim)

    return n_pairs


#the logistic eqxn
def _calc_logistic_growth(R, N, K):
    dNdt = R*(1-(N/K))*N
    return dNdt


def _calc_logistic_soln(x0, R, t):
        return 1/(1+((1/x0)-1)*np.e**(-1*R*t))


def _calc_dNdt(R, N, K, pop_growth_eq = 'logistic'):
    with np.errstate(divide='ignore', invalid='ignore'):
        #if using the logistic equation for pop growth
        #NOTE: For now this is the only option and the default, but could 
        #easily offer other equations later if desired
        if pop_growth_eq == 'logistic':
            #use logistic eqxn, with spp intrinsic growth rate (spp.R)
            #to generate current growth-rate raster
            dNdt = _calc_logistic_growth(R, N, K)
    #coerce extreme negative values, NaNs, and infs to a reasonable value
    #(extreme negative values are produced in areas where the K raster
    #has extremely small values)
    dNdt = np.clip(dNdt, a_min = -1*N.max(), a_max = None)
    dNdt[np.isnan(dNdt)] = -1*N.max()
    dNdt[np.isinf(dNdt)] = -1*N.max()
    return dNdt


def _calc_N_b(b, n_births_distr_lambda, n_pairs):
    #Use n_pairs, spp.b (i.e. probability that a pair gives birth),
    #and spp.n_births_distr_lambda (i.e. expected number of births
    #per pair) to calculate the N_b raster (expected number of births in each
    #cell) according to N_b = b*E[N_b/pair]*n_pairs where E[N_b/pair] = 
    #n_births_distr_lambda (b/c births Poission distributed)
    #NOTE: Would be good to in the future to use neg_binom dist to allow
    #births to be modeled as Poisson (using correct parameterization of
    #neg_binom to approximate that) or overdispersed (to be parmeterized
    #with field data)

    #NOTE: Does it matter that I'm not attempting to account for the fact
    #that pairs will disperse out of the cell? Almost certainly a larger effect
    #with increasing dispersal-kernel distance. But nonetheless, would it even
    #make sense (and even BE TRACTABLE) to calculate it any other way? My
    #intuition is NO, but worth considering further later on...

    #NOTE: Currently, birth rate is intrinsic to the species and cannot be
    #made a function of density or spatial fitness. Might be worth adding
    #this at some point in the future though.
    N_b = b * n_births_distr_lambda * n_pairs
    return N_b


def _calc_Nd(N_b, dNdt):
    #Use N_b - dN/dt to calculate N_d raster (expected number of deaths in
    #each cell)
    N_d = N_b - dNdt
    return N_d


def _calc_d(N_d, N, d_min, d_max):
    #Use N_d/N to calculate d raster (probability of death of individuals
    #in each cell) Calculate d as simply N_d/N, then control for some
    #numerical artefacts that an inappropriate
    # NOTE: ignore x/0. and 0./0. warnings
    with np.errstate(divide='ignore', invalid='ignore'):
        d = N_d/N
    #fix infinties and NaNs and negatives if they arise
    #(they occur where N ==0 or where N_d == 0)
    d[np.isnan(d)] = 0 
    #constrain to the min and max d values
    d = np.clip(a = d, a_min = d_min, a_max = d_max)

    #NOTE: DEH: 12-08-18: Got rid of the infinity check, because it was
    #coercing all infinities to 0, but in many cases infinities should actually
    #have been 1s, so it created numerous problems that just clipping to d_min
    #and d_max takes care of anyhow; commented out the NaN check below, but not
    #entirely clear that I don't need it...
    #d[np.isnan(d)] = 0
    return d


def _do_mortality(spp, death_probs):
    deaths = np.array([*spp])[np.bool8(r.binomial(n = 1, p = death_probs))]
    if len(deaths) > 0:
        ig = itemgetter(*deaths)
        [spp.pop(ind) for ind in deaths];
    return len(deaths)


def _do_pop_dynamics(spp, land, with_selection = True, burn = False,
    births_before_deaths = False, asserts = True, debug = False):
    '''Generalized function for implementation population dynamics.
    Will carry out one round of mating and death, according to parameterization
    laid out in params dict (which were grabbed as Species attributes).

       If with_selection == False, only density-dependent death will occur.

       If burn == True, selection will be coerced to False, the Species.K
       attribute will be updated, stationarity metrics will be assessed, and
       the function will return the decision it has reached regarding
       stationarity.

    '''

    #NOTE: Would be good to write in option to ONLY mate or ONLY die when the
    #function is run (so that a spp could be run with mating every x timesteps
    #but mortality each timestep, for example)

    #create the debug-plotter, id debug == True
    if debug:
        timestep = -99999
        dp = _DebugPlotter(land, spp, timestep)

    #find mating pairs
    pairs = spp._find_mating_pairs()

    #calc num_pairs raster (use the calc_pop_density function on the
    #midpoints of the mating pairs)
    n_pairs = _calc_n_pairs(pairs = pairs, spp = spp)
    #run checks on n_pairs
    if asserts:
        assert n_pairs.min() >= 0, 'n_pairs.min() == %0.2f' %(n_pairs.min())
        assert not np.any(np.isnan(n_pairs))
        assert not np.any(np.isinf(n_pairs))
    #add debug plot
    if debug:
        dp._next_plot('n_pairs', n_pairs)

    #if births should happen before (and thus be included in the calculation of)
    #deaths, then mate and disperse babies now
    if births_before_deaths:
        #Feed the land and mating pairs to spp.do_mating, to produce and
        #disperse zygotes
        spp._do_mating(land, pairs, burn)

    #calc N raster, set it as spp.N, then get it  
    spp._calc_density(set_N = True)
    N = spp.N
    #run checks on N
    if asserts:
        assert N.min() >= 0
        assert not np.any(np.isnan(N))
        assert not np.any(np.isinf(N))
    #add debug plot
    if debug:
        dp._next_plot('N', N)

    #get K raster
    K = spp.K
    #run checks on K
    if asserts:
        assert K.min() >= 0
        assert not np.any(np.isnan(K))
        assert not np.any(np.isinf(K))
    #add debug plot
    if debug:
        dp._next_plot('K', K)

    #calc dNdt
    dNdt = _calc_dNdt(R = spp.R, N = N, K = K,
                                    pop_growth_eq = 'logistic')
    #run checks on dNdt
    if asserts:
        assert not np.any(np.isnan(dNdt))
        assert not np.any(np.isinf(dNdt)), ('The following cells are '
            'infinite: \n\t%s') % str([i for i, n in enumerate(
                                        dNdt.ravel()) if np.isinf(n)])
    #add debug plot
    if debug:
        dp._next_plot('dNdt', dNdt)

    #calculate N_b (raster of estimated births)
    N_b = _calc_N_b(b = spp.b,
        n_births_distr_lambda = spp.n_births_distr_lambda, n_pairs = n_pairs)
    #run checks on N_b
    if asserts:
        assert N_b.min() >= 0
        assert not np.any(np.isnan(N_b))
        assert not np.any(np.isinf(N_b))
    #add debug plot
    if debug:
        dp._next_plot('N_b', N_b)

    #calc N_d (raster of deaths)
    N_d = _calc_Nd(N_b = N_b, dNdt = dNdt)
    #run checks on N_d
    if asserts:
        assert not np.any(np.isnan(N_d))
        assert not np.any(np.isinf(N_d))
    #add debug plot
    if debug:
        dp._next_plot('N_d', N_d)

    #calc d (raster of probabilities of density-dependent death)
    d = _calc_d(N_d = N_d, N = N, d_min = spp.d_min, d_max = spp.d_max)

    #run checks on d
    if asserts:
        assert d.min() >= 0, 'd.min() is %0.2f, at %s' % (d.min(),
                                                        str(d.argmin()))
        assert d.max() <= 1, 'd.max() is %0.2f' % d.max()
        assert not np.any(np.isnan(d))
        assert not np.any(np.isinf(d))
    #add debug plot
    if debug:
        dp._next_plot('d', d)

    #If births should happen after (and thus not be included in the 
    #calculation of) deaths, then instead of having mated and dispersed
    #babies up above, do it now (i.e. now that the d raster has been 
    #calculated)
    #NOTE: The way I'm looking at it now, calculating likely births from number
    #of existing pairs, then calculating death raster, then having births take
    #place, then having per-individual death probabilies determined and having
    #deaths take place (such that newborns could also immediately die after
    #birth, based on the death raster calculated before they were born, and
    #even based on individs who weren't their parents, assuming they dispersed
    #across cell lines), which is what I'm doing currently, is the only
    #obvious/reasonable way to do this. Of course there will always be an
    #inherent rub between reality and discrete-time simulations because of
    #the continuous/discrete time conflict. But in this case, is there any
    #strong reason to CHANGE THIS? Or are there any foreseeable and undesirable
    #results/effects?... NEED TO PUT MORE THOUGHT INTO THIS LATER.
    if not births_before_deaths:
        #Feed the land and mating pairs to the mating functions, to produce
        #and disperse zygotes
        spp._do_mating(land, pairs, burn)
    #Get death probabilities
    death_probs = d[spp.cells[:,1], spp.cells[:,0]]
    #If with_selection (i.e. if death probs should account for fitness),
    #then use the d raster and individuals' fitnesses to calculate
    #per-individual probabilities of death
    if with_selection:
        death_probs = _calc_prob_death(spp, death_probs)
    #run checks on death_probs
    if asserts:
        assert np.alltrue(death_probs >= 0)
        assert np.alltrue(death_probs <= 1)

    #kill (and track kills) based on max_age
    num_killed_age = 0
    if spp.max_age is not None:
        death_probs[spp._get_age() > spp.max_age] = 1
        num_killed_age = np.sum(death_probs == 1)

    #Use the per-individual death probabilities to carry out mortality 
    num_deaths = _do_mortality(spp, death_probs)
    spp._set_coords_and_cells()
    spp.n_deaths.append(num_deaths)

    #Check if extinct, and return result
    extinct = spp._check_extinct()
    return(extinct)





#!/usr/bin/python
#_memory_help.py

'''
Helper functions for assesing the memory usage of Geonomics objects.
'''

#------------------------------------
# CLASSES ---------------------------
#------------------------------------


#--------------------------------------
# FUNCTIONS ---------------------------
#--------------------------------------

#A nice function for recursively iterate and sum the size of an object
#Courtesy of Aaron Hall's response at
#http://stackoverflow.com/questions/449560/
#how-do-i-determine-the-size-of-an-object-in-python
def get_obj_size(obj_0):
    """
    Get the full size of a complex object

    Recursively iterate through the structure of some complex object,
    summing the size of all components, to return the full size of the object.

    Parameters
    ----------
    obj_0, object
        The object whose full size should be calculated.

    Returns
    -------
    out : float
        The full calculated size of the the input object.
    """

    import sys
    from numbers import Number
    from collections import Set, Mapping, deque

    try: # Python 2
        zero_depth_bases = (basestring, Number, xrange, bytearray)
        iteritems = 'iteritems'
    except NameError: # Python 3
        zero_depth_bases = (str, bytes, Number, range, bytearray)
        iteritems = 'items'

    def inner(obj, _seen_ids = set()):
        obj_id = id(obj)
        if obj_id in _seen_ids:
            return 0
        _seen_ids.add(obj_id)
        size = sys.getsizeof(obj)
        if isinstance(obj, zero_depth_bases):
            pass # bypass remaining control flow and return
        elif isinstance(obj, (tuple, list, Set, deque)):
            size += sum(inner(i) for i in obj)
        elif isinstance(obj, Mapping) or hasattr(obj, iteritems):
            size += sum(inner(k) + inner(v) for k, v in getattr(obj,
                iteritems)())
        # Check for custom object instances - may subclass above too
        if hasattr(obj, '__dict__'):
            size += inner(vars(obj))
        if hasattr(obj, '__slots__'): # can have __slots__ with __dict__
            size += sum(inner(getattr(obj,
                s)) for s in obj.__slots__ if hasattr(obj, s))
        return size
    return inner(obj_0)




#!/usr/bin/python
#stats.py


'''
Classes and functons to implement calculation and output of statistics
'''

#geonomics imports
from geonomics.utils.io import (_append_array2d_to_array_stack,
                                _append_row_to_csv, _write_dict_to_csv)
from geonomics.ops.selection import _calc_fitness
from geonomics.utils.viz import _check_display

#other imports
import numpy as np
from scipy.stats.stats import pearsonr
from collections import Counter as C
import os
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

#a StatsCollector class, to parameterize and manage calculation
#and collection of stats, then write them to file at the end of 
#each model iteration
class _StatsCollector:
    def __init__(self, model_name, params):

        #set model_name
        self.model_name = model_name

        #set total model time
        self.T = params.model.T

        #grab the stats parameters
        stats_params = params.model.stats

        #a dictionary to link the stats' names in the params dict 
        #to the functions to be called to calculate them 
        self.calc_fn_dict = {'Nt': _calc_Nt,
                             'ld':  _calc_ld,
                             'het': _calc_het,
                             'maf': _calc_maf,
                             'mean_fit': _calc_mean_fitness,
                              }

        #a dictionary to link the stats' names in the params dict 
        #to the functions to be called to write them to disk
        self.write_fn_dict = {'ld':  self._write_array_to_stack,
                              'het': self._write_row_to_csv,
                              'maf': self._write_row_to_csv,
                             }

        #a dictionary to link stats to the file extensions that 
        #should be used to write them to disk
        self.file_suffix_dict = {'Nt': 'OTHER_STATS.csv',
                                 'ld':  'LD.txt',
                                 'het': 'HET.csv',
                                 'maf': 'MAF.csv',
                                 'mean_fit': 'OTHER_STATS.csv',
                                 }

        #get the species names
        spps_with_wout_genomes = {str(k):('gen_arch' in v.keys()) for k, v
                                 in params.comm.species.items()}

        #list stats that cannot be calculated for species without genomes
        stats_invalid_wout_genomes = ['ld', 'het', 'maf', 'mean_fit']

        #create a stats attribute, to store all stats calculated
        self.stats = {}
        for spp_name, genome in spps_with_wout_genomes.items():
            self.stats[spp_name] = {}
            for stat, stat_params in stats_params.items():
                #skip species without genomes for stats that need genomes
                if not genome and stat in stats_invalid_wout_genomes:
                    break
                #each spp gets a subdict
                else:
                    #each subdict gets a key for each stat to be calculated
                    if stat_params.calc:
                        #create a subdictionary for each stat, with a list of
                        #NaNs self.T items long, which will be filled in for
                        #each whenever it is sampled (NOTE: this forces all
                        #stats to have the same length so that they all fit
                        #into one pd.DataFrame at the end, and so that plots
                        #easily line up on the same timeframe)
                        self.stats[spp_name][stat]= {
                            'vals': [np.nan]*self.T,
                            'freq': stat_params.freq,
                            #add a 'filepath' key, whose value will be updated
                            #to contain to correct filepaths for each stat
                            'filepath': None,
                            #create tuple of other, stat-specific parameters, 
                            #to later be unpacked as arguments to
                            #the appropriate stat function
                            'other_params': dict([(k,v) for k,v in
                              stat_params.items() if k not in ['calc',
                                                                    'freq']])
                            }

                        #if the freq value is 0, change it to self.T -1, so
                        #that it collects only on the first and last timestep
                        if self.stats[spp_name][stat]['freq'] == 0:
                            self.stats[spp_name][stat]['freq'] = self.T-1

    #create a master method, to be called each timestep, which will make a list 
    #of all stats that need to be calculated that timestep (based on the 
    #calculation-frequencies provided in the params dicts), and then calls the
    #functions to calculate them all and adds the results to self.stats
    def _calc_stats(self, community, t, iteration):
        #set the filepaths, if this is the first timestep of the model
        #iteration
        if t == 0:
            self._set_filepaths(iteration)
        #for each species
        for spp in community.values():
            #list the stats to be calculated this timestep
            if t == self.T-1:
                #calculate all, if it's the last timestep
                calc_list = [*self.stats[spp.name]]
            else:
                #or else only calculate based on the parameterized frequencies
                #for each stat
                calc_list = [k for k,v in self.stats[spp.name].items() if (
                                                        t % v['freq'] == 0)]
            #then calculate each stat
            for stat in calc_list:
                vals = self.calc_fn_dict[stat](spp,
                                **self.stats[spp.name][stat]['other_params'])
                #and add each stat to the right location (by timestep)
                #in its list
                try:
                    self.stats[spp.name][stat]['vals'][t] = vals
                #unless the list isn't long enough (which happens if mod.walk
                #has been used to run the model past its initially stipulated
                #length of time), in which case make it long enough and make
                #the last value the stat just calculated
                except IndexError:
                    stats_list = self.stats[spp.name][stat]['vals']
                    stats_list.extend([np.nan] * (t-len(stats_list)) + [vals])
        #and write whichever stats are necessary to file
        self._write_stats(t)

    #a method to make the filenames for all of the stats to be saved
    def _set_filepaths(self, iteration):
        #get the directory name for this model and iteration
        dirname = os.path.join('GNX_mod-%s' % self.model_name,
                               'it-%i' % iteration)
        #for each species
        for spp_name in [*self.stats]:
            #get the subdirectory name and filename for this species
            subdirname = os.path.join(dirname, 'spp-%s' % spp_name)
            #make this subdir, and any parent dirs as necessary
            os.makedirs(subdirname, exist_ok = True)
            #create the filename and filepath for this spp, for each stat
            for stat in [*self.stats[spp_name]]:
                filename = 'mod-%s_it-%i_spp-%s_%s' % (self.model_name,
                    iteration, spp_name, self.file_suffix_dict[stat])
                filepath = os.path.join(subdirname, filename)
                #add the filepath for this stat to self.stats
                self.stats[spp_name][stat]['filepath'] = filepath

    #wrapper around io.append_array2d_to_array_stack
    #TODO WHAT TO DO WITH t IN THIS CASE?? CAN'T ADD TO txt 3D ARRAY FILE
    def _write_array_to_stack(self, filepath, array, t):
        _append_array2d_to_array_stack(filepath, array)

    #wrapper around io.append_row_to_csv
    def _write_row_to_csv(self, filepath, array, t):
        _append_row_to_csv(filepath, array, t)

    #use io._write_dict_to_csv to write to disk all "other stats", i.e. 
    #all stats that collect only a single value per species per timestep
    #TODO: CHANGE THE 'OTHER STATS' NAMING CONVENTION TO SOMETING MORE
    #DESCRIPTIVE
    def _write_other_stats(self):
        for spp, spp_stats in self.stats.items():
            #get a dictionary of the data values for all stats that are to be
            #written just once at the end of the iteration
            data_dict = {k:v['vals'] for k,v in spp_stats.items() if
                              'OTHER_STATS' in v['filepath']}
            #they all have the same filepath, so just grab the first
            filepath = [*spp_stats.values()][0]['filepath']
            #write to disk
            _write_dict_to_csv(filepath, data_dict)

    #method to write stats to files, in the appropriate directory (by model
    #and iteration number), and with the appropriate spp names in the filenames
    def _write_stats(self, t):
        #for each species
        for spp_name, spp_stats in self.stats.items():
            #for each stat
            write_list = [k for k,v in spp_stats.items() if t % v['freq'] == 0]
            for stat, stat_dict in spp_stats.items():
                #get the filepath
                filepath = stat_dict['filepath']
                #if the filepath does not contain "OTHER_STATS" then it is a
                #stat that produces more than a single value per species per
                #timestep it is collected, so write the data to disk
                #intermittently and then delete the data from memory (if it was
                #collected this timestep)
                if "OTHER_STATS" not in filepath and stat in write_list:
                    #get the correct write_fn for this stat
                    write_fn = self.write_fn_dict[stat]
                    #call the write_fn to write the data to disk
                    write_fn(filepath, stat_dict['vals'][t], t)
                    #then replace the last data collected prior to this
                    #timestep's data with None, to free up memory but still
                    #maintain the latest data in case of plotting
                    rev_nonnull = [n for n, v in enumerate(
                        stat_dict['vals'][::-1]) if (v is not np.nan and
                                                        v is not None)]
                    nonnull = [range(len(
                        stat_dict['vals']))[::-1][n] for n in rev_nonnull]
                    nonnull = [v for v in nonnull if v != t]
                    for v in nonnull:
                        stat_dict['vals'][v] = None


        #or write all 'other stats' to disk, if it's the last timestep
        if t == self.T-1:
            self._write_other_stats()

    #method to plot whichever stat as a function of runtime
    def _plot_stat(self, stat, spp_name=None):
        #check that the stat argument is valid
        assert type(stat) is str, "The 'stat' argument must be a string."
        assert stat in [*self.stats.values()][0].keys(), ("The 'stat' "
            "argument must name a statistic that was collected. Valid values: "
            "%s.") % (','.join(["'%s'" % val for val in
                                [*self.stats.values()][0].keys()]))
        #get the list of spps to plot
        if spp_name is None:
            spp_names = [*self.stats]
        elif (spp_name is not None
              and type(spp_name) is str and spp_name in [*self.stats]):
            spp_names = [spp_name]
        else:
            raise ValueError(("The 'spp_name' argument, if provided, "
                "must be a string containing a valid species name."))
        #create the figure
        fig = plt.figure()
        #plot each species for the chosen statistic 
        for n, spp_name in enumerate(spp_names):
            #get the stat values to plot
            vals = self.stats[spp_name][stat]['vals']
            #plot 'Nt' or 'mean_fit'
            if stat in ['Nt', 'mean_fit']:
                #add axes objects horizontally across
                ax = fig.add_subplot(1, len(spp_names), n+1)
                #get the indices of non-NaN values to be plotted
                indices_to_plot = np.array(np.where(
                                    np.invert(np.isnan(vals)))[0])
                #get the timesteps at the non-NaN values
                x = np.arange(0, len(vals))[indices_to_plot]
                #get the non-NaN values
                y = np.array(vals)[indices_to_plot]
                #plot a dotted line (which necessarily linearly interpolates 
                #between collected timesteps if not all timesteps
                #were collected)
                plt.plot(x, y, ':')
                #and plot dots at each of the collected timesteps
                plt.plot(x, y, '.')
                #set the title to the stat and the species' name
                ax.set_title("SPP: '%s'" % (spp_name))
                #set the x- and y-labels
                plt.xlabel('timestep')
                plt.ylabel(stat)

            #or plot 'maf' or 'het'
            elif stat in ['het', 'maf']:
                #add axes objects horizontally across
                ax = fig.add_subplot(1, len(spp_names), n+1)
                #get the reversed-list index of the last set of values 
                #calculated
                rev_idx_last_vals = [n for n,v in enumerate(vals[::-1]) if (
                                    v is not None and v is not np.nan)][0]
                #get the last set of values calculated
                last_vals = vals[::-1][rev_idx_last_vals]
                #get the timestep of the last set of values
                t_last_vals = range(len(vals))[::-1][rev_idx_last_vals]
                #plot the values
                plt.plot(range(len(last_vals)), last_vals, '-')
                #set the title to the species' name and timestep of the
                #values plotted
                ax.set_title("SPP: '%s';   T: %i" % (spp_name, t_last_vals))
                #set the x- and y-labels
                plt.xlabel('locus')
                plt.ylabel(stat)

            #or plot 'ld'
            elif stat in ['ld']:
                #get the reversed-list index of the last set of values 
                #calculated
                rev_idx_last_vals = [n for n,v in enumerate(vals[::-1]) if (
                    v is not None and v is not np.nan)][0]
                #get the last set of values (i.e. r^2 array) calculated
                r2_mat = vals[::-1][rev_idx_last_vals]
                #get the timestep of the last set of values
                t_last_vals = range(len(vals))[::-1][rev_idx_last_vals]
                #add axes objects horizontally across, in two rows
                ax = fig.add_subplot(2, len(spp_names), n+1)
                #plot the LD matrix in row 1
                plt.imshow(np.clip(r2_mat, a_min = 0, a_max = None),
                                            interpolation = 'nearest')
                plt.colorbar()
                #set plot title
                ax.set_title(("SPP: '%s';   T: %i\nLocus-wise "
                            "linkage matrix") %  (spp_name, t_last_vals))
                #set the x- and y-labels
                plt.xlabel('locus')
                plt.ylabel('locus')
                ax = fig.add_subplot(2, len(spp_names), n+1+len(spp_names))
                #plot of mean linkage values
                r2_list = [r2_mat[0,1]]
                L = r2_mat.shape[0]
                for i in range(1,L-1):
                    r2_list.append(np.mean([r2_mat[i-1,i], r2_mat[i,i+1]]))
                r2_list.append(r2_mat[L-2,L-1])
                plt.scatter(range(L), r2_list, c = 'red', marker = 'o', s=25)
                #set plot title
                ax.set_title("Locus-wise mean linkage values")
                #set the x- and y-labels
                plt.xlabel('locus')
                plt.ylabel('mean linkage')

            #or else return informative error message
            else:
                raise ValueError(("The value provided for the 'stat' argument "
                    "is not a valid statistic. Valid values include: %s\n\n")%(
                    ','.join(['%s' % k for k in [*self.calc_fn_dict]])))
        #set the main title to the stat plotted
        fig.suptitle('STAT: %s' % stat)
        #show the image
        fig.show()


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#method to get pop size (NOTE: not actually calculating it)
def _calc_Nt(spp):
    Nt = spp.Nt[-1]
    return(Nt)


def _calc_ld(spp, plot = False):

    #TODO: I should also include (either as an alternative within this fn,
    #or as separate fn) the option to calculate D'

    #TODO: I keep getting warnings like the following, which could just be 
    #due to divison of small floating-point numbers, but I should figure out 
    #exactly what's going on and be sure everything checks out. WARNING:
    # stats.py:117: RuntimeWarning: invalid value encountered in double_scalars

    speciome = spp._get_genotypes()
    n = np.shape(speciome)[0] #num individs
    x = np.shape(speciome)[2] #ploidy
    N = n*x
    L = spp.gen_arch.L
    assert L == np.shape(speciome)[1], ("The length of the 1st dimension "
                            "of speciome doesn't equal spp.genomic_arch.L")

    r2_mat = np.zeros([L]*2) * np.nan # vals default to NaN

    for i in range(L):
        for j in range(i+1, L):
            #calculate freq of allele 1 at locus i
            f1_i = np.sum(speciome[:,i,:], axis = None)/(N)
            #calculate freq of allele 1 at locus j
            f1_j = np.sum(speciome[:,j,:], axis = None)/(N)
            #calculate freq of chroms with 1_1 haplotype at loci i and j
            f11_ij = float(np.sum(speciome[:,[i,j],:].sum(axis = 1) ==2,
                                                        axis = None))/(N)
            D_1_1 = f11_ij - (f1_i * f1_j)
            r2 = (D_1_1**2)/(f1_i*(1-f1_i)*f1_j*(1-f1_j))
            # add to both triangular halves, to produce a symmetric matrix
            r2_mat[i,j] = r2
            r2_mat[j,i] = r2

    return(r2_mat)


#function to calculate the locus-wise (if mean == False) or mean (if
#mean == True) heterozygosity of the species
def _calc_het(spp, mean=False):
    #get pop size
    N = len(spp)
    #get the speciome
    speciome = spp._get_genotypes()
    #calculate the frequency of heterozygotes, locus-wise
    het = np.sum(np.mean(speciome, axis = 2) == 0.5, axis = 0)/N
    #get the mean heterozygosity, if mean argument is True
    if mean:
        het = mean(het)
    return(het)

#function to calculate the locus-wise minor allele frequency of the species
def _calc_maf(spp):
    #get two times the pop size
    two_N = 2*len(spp)
    #get the speciome
    speciome = spp._get_genotypes()
    #get the frequencies of 1-alleles for all loci
    freqs_1 = np.sum(np.sum(speciome, axis = 2), axis = 0)/two_N
    #find all loci where the 1-allele is the major allele
    majors = np.where(freqs_1 > 0.5)
    #replace the locations where 1 is the major allele with 0-allele freq
    maf = freqs_1[:]
    maf[majors] = 1 - freqs_1[majors]
    return(maf)


#function to calculate the mean fitness of the species
def _calc_mean_fitness(spp):
    #calculate the mean fitness, if this species has traits
    if spp.gen_arch.traits is not None:
        mean_fit = np.mean(_calc_fitness(spp))
    #or else return NaN
    else:
        mean_fit = np.nan
    return(mean_fit)



#!/usr/bin/python
#mutation.py

'''
Functions to implement mutation operations
'''

#other imports
import numpy as np
from numpy import random as r
import random
import re


#------------------------------------
# CLASSES ---------------------------
#------------------------------------


#--------------------------------------
# FUNCTIONS ---------------------------
#--------------------------------------

def _calc_estimated_total_mutations(spp, burn_T, T):
    #NOTE: this would actually be a pretty poor estimate, because mutations
    #will occur in new individuals, not some static population
    mean_births = np.mean(spp.n_births[-burn_T:])
    est = mean_births * spp.gen_arch.L * T * spp.gen_arch._mu_tot
    #give a decent overestimate
    est = int(2.5 * est)
    return est

# add a row to the tskit.TableCollection.mutations table for a new mutation
def _do_add_row_muts_table(spp, individ, homol, locus):
    # update the tskit.TableCollection.mutations table
    node_id = spp[individ]._nodes_tab_ids[homol]
    mut_id = spp._tc.mutations.add_row(site=locus, node=node_id,
                                       derived_state='1')
    return mut_id


# do a neutral mutation for a single Individual, chosen from the offspring
def _do_neutral_mutation(spp, offspring, locus=None, individ=None):
    #randomly choose an individual to mutate, if not provided
    if individ is None:
        individ = r.choice(offspring)
    else:
        assert individ in spp.keys(), ('The individual provided '
                                       'is not in spp.keys().')
    #randomly choose a locus from among the mutables, if not provided
    if locus is None:
        locus = spp.gen_arch._mutables.pop()
    else:
        assert locus in spp.gen_arch._mutables, ('The locus provided '
                                                 'is not in the list of '
                                                 'valid mutable loci '
                                                 '(spp.gen_arch._mutables).')
    #randomly choose a homologue on which to place the mutation
    #in the individual
    homol = r.binomial(1,0.5)
    # NOTE: not mutating the individual's genome because we're only tracking
    # non-neutral mutations now

    # add a row to the tskit.TableCollection.mutations table, if necessary
    if spp.gen_arch.use_tskit:
        _do_add_row_muts_table(spp, individ, homol, locus)
    return(individ, locus)


# do a non-neutral mutation for a single Individual, chosen from the offspring
def _do_nonneutral_mutation(spp, offspring, locus=None, individ=None,
                           trait_nums=None, delet_s=None):
    # choose a new locus, if not provided
    if locus is None:
        locus = spp.gen_arch._mutables.pop()
    # otherwise, remove from the mutables the locus that was provided
    else:
        spp.gen_arch._mutables.remove(locus)
    # draw an individual to mutate from the list of offspring provided,
    # unless individ is provided
    if individ is None:
        individ = r.choice(offspring)
    else:
        assert individ in offspring, ('Individual provided for mutation (%i) '
                                      'does not appear to be among the '
                                      'list of current-timestep '
                                      'offspring (%s)') % (individ,
                                                           str(offspring))
    # add the locus to the gen_arch.nonneut_loci array
    idx = spp.gen_arch._add_nonneut_locus(locus, trait_nums, delet_s)
    #create the mutation
    homol = r.binomial(1, 0.5)

    # add a new row for this locus, with '0' genotypes,
    # to the individuals' genotype arrays,
    # NOTE: only is using tskit, because only in that case will individuals'
    #       genotype arrays not already contain all simulated loci
    if spp.gen_arch.use_tskit:
        spp._add_new_locus(idx, locus)

    # mutate the chosen individual's chosen homologue's genotype to 1
    spp[individ].g[idx, homol] = 1
    # update the individual's phenotype
    spp._set_z_individ(individ)

    # manage tskit data structs, if necessary
    if spp.gen_arch.use_tskit:
        # add a row to the tskit.TableCollection.mutations table
        _do_add_row_muts_table(spp, individ, homol, locus)
        # update the recombination subsetters
        spp.gen_arch.recombinations._update_subsetters(locus, idx)
    return(individ, locus)


# do a trait mutation for a single Individual, chosen from the offspring
def _do_trait_mutation(spp, offspring, trait_nums, alpha=None,
                                        locus=None, individ=None):
    #run the do_nonneutral_mutation function, to select locus
    #and individ (unless already provided) and update
    #the mutable_loci, neut_loci, and nonneut_loci, and 
    #change one of this locus' alleles to 1 in the mutated individual
    individ, locus = _do_nonneutral_mutation(spp=spp, offspring=offspring,
                                             locus=locus, individ=individ,
                                             trait_nums=trait_nums)
    return(individ, locus)


# do a deleterious mutation for a single Individual, chosen from the offspring
def _do_deleterious_mutation(spp, offspring, locus=None, s=None, individ=None):
    #choose a selection coefficient for this locus, if not provided
    if s is None:
        s = spp.gen_arch._draw_delet_s()
    #run the do_nonneutral_mutation function, to select locus and individ
    #(unless already provided) and update the mutable_loci, neut_loci, and 
    #nonneut_loci sets, and change one of this locus' alleles to 1 in
    #the mutated individual
    individ, locus = _do_nonneutral_mutation(spp=spp, offspring=offspring,
                                             locus=locus, individ=individ,
                                             delet_s=s)
    #return the locus and individual
    return(individ, locus)


#TODO: COMPLETE THIS?
def _do_planned_mutation(planned_mut_params):
    pass


# do mutations for a list of offspring
def _do_mutation(offspring, spp, log=None):
    #draw number of mutations from a binomial trial with number of trials equal
    #to len(offspring)*spp.gen_arch.L and prob = sum(all mutation rates)
    n_muts = r.binomial(n = len(offspring) * spp.gen_arch.L,
                                    p = spp.gen_arch._mu_tot)
    #TODO: ADD A PRINT STATEMENT FOR 'verbose' MODE INSTEAD
    #if n_muts > 0:
    #   print('\t**** %i mutations occurred\n\n' % n_muts)

    #if n_muts >0, draw which type of mutation
    #each resulting mutation should be (weighted by relative mutation rates)
    if n_muts > 0:
        muts = spp.gen_arch._draw_mut_types(num = n_muts)

        #for each mutation, add the corresponding mutation function to a queue
        mut_queue = []
        for mut in muts:
            #add the appropriate mut_fn to the queue
            mut_fn = spp.gen_arch._mut_fns[mut]
            mut_queue.append(mut_fn)

        #add to the queue any planned mutations, if necessary
        if spp.gen_arch._planned_muts:
            if spp.gen_arch._planned_muts.next[0] <= spp.t:
                mut_queue.append(spp.gen_arch._planned_muts.next[1])
                spp.gen_arch._planned_muts._set_next()

        #execute all functions in the queue
        for i, fn in enumerate(mut_queue):
            individ, locus = fn(spp, offspring)
            log_msg = ('MUTATION: %s\n\t INDIVIDUAL %i,  '
                'LOCUS %i\n\t timestep %i\n\n') % (muts[0], individ,
                                                   locus, spp.t)
            if log:
                with open(log, 'a') as f:
                    f.write(log_msg)
            print(log_msg)
            mut_loci_exhausted = False



#!/usr/bin/python
#selection.py


'''
Functions to implement natural selection operations (aside from actual
mortality)
'''


import numpy as np
from collections import OrderedDict as OD


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#Get the phenotypic values of all individuals for a given trait
def _calc_phenotype(ind, gen_arch, trait_num):
    # get number of loci and their allelic effect sizes
    n_loci = gen_arch.traits[trait_num].n_loci
    alpha = gen_arch.traits[trait_num].alpha
    # get the mean genotype array (using the trait's locus index)
    # NOTE: use loci_idxs if tskit is being used and Individuals' genotype
    #       arrays only hold their non-neutral genotypes;
    #       otherwise, just use the trait.loci directly
    if gen_arch.use_tskit:
        genotype = np.mean(ind.g[gen_arch.traits[trait_num].loci_idxs], axis=1)
    else:
        genotype = np.mean(ind.g[gen_arch.traits[trait_num].loci], axis=1)
    #use dominance, if required (to save considerable compute time otherwise)
    if gen_arch._use_dom:
        #get the dominance values
        dom = gen_arch.dom[gen_arch.traits[trait_num].loci]
        #update the genotype array by accounting for dominance at each locus
        genotype = np.clip(genotype * (1 + dom), a_min = None, a_max = 1)
    #if polygenic, multiply genotype by alpha (i.e. effect size) 
    #at all loci, sum across loci, then add to 0.5 (the null phenotypic
    #value), to get phenotype
    if n_loci > 1:
        phenotype = 0.5 + sum(genotype*alpha)
    #else if monogenic, then mean genotype = phenotype
    else:
        phenotype = genotype[0]
    return(phenotype)


def _calc_fitness_one_trait(t, e, z, spp):
    fit = 1 - t._get_phi(spp)*(abs((
                e[:,t.lyr_num]**(not t.univ_adv)) - z[:,t.idx])**t.gamma)
    return(fit)


def _calc_fitness_traits(spp, trait_num = None):
    traits = spp.gen_arch.traits.values()
    #subset for single trait, if indicated
    if trait_num is not None:
        traits = [list(traits)[trait_num]]
    #get all individuals' environmental values
    e = spp._get_e()
    #get all individuals' phenotypes
    z = spp._get_z()
    #create lambda function with current e, z, and spp objects
    calc_fitness_lambda = lambda t: _calc_fitness_one_trait(t, e, z, spp)
    #map the calc_sngl_trait_fitness function to all traits, then
    #calculate overall fitness as product of
    #fitness for each trait
    fit = np.stack(list(map(calc_fitness_lambda, traits))).prod(axis = 0)
    #for polygenic traits, loci with very large effect sizes can 
    #generate fitnesses less than 0; correct for this
    fit = np.clip(fit, a_min = 0.001, a_max = None)
    return(fit)


def _calc_fitness_deleterious_mutations(spp):
    #create an np.array that has individuals in rows and their
    #diploid genotypes for each of the deleterious loci in the cols
    #(0, 1, or 2, to facilitate the fitness math, because s values
    #(i.e. selection coefficients) are expressed per allele)
    # NOTE: use delet_loci_idxs if tskit is being used and Individuals' genotype
    #       arrays only hold their non-neutral genotypes;
    #       otherwise, just use the delet_loci directly
    if spp.gen_arch.use_tskit:
        deletome = np.sum(np.stack([ind.g[
            spp.gen_arch.delet_loci_idxs, :] for ind in spp.values()]), axis=2)
    else:
        deletome = np.sum(np.stack([ind.g[
            spp.gen_arch.delet_loci, :] for ind in spp.values()]), axis=2)
    fit = 1 - np.multiply(deletome, spp.gen_arch.delet_loci_s)
    fit = fit.prod(axis = 1)
    return(fit)


#one function to calculate total fitness, including traits and deleterious
#loci, as applicable
def _calc_fitness(spp, trait_num=None):
    #set a default w array
    w = np.array([1]*len(spp))
    #get trait-related fitness, if traits
    if (spp.gen_arch.traits is not None
        and len(spp.gen_arch.traits) > 0):
        w = w * _calc_fitness_traits(spp, trait_num = trait_num)
    #if fitness is not supposed to be calculated for a specific trait, and if 
    #species has deleterious mutations, then get the fitnesses related to 
    #the deleterious traits (i.e. operationalize background selection)
    if (trait_num is None and
        len(spp.gen_arch.delet_loci) > 0):
            w = w * _calc_fitness_deleterious_mutations(spp)
    return w


#Get the vector of mortalities (probabilies of death) for a 
#given density-dependent Pr(death) at a cell, the environmental value(s)
#at that cell, the phenotype(s) of the trait(s) for the individuals found
#there, and the selection coefficient(s) on the trait(s)
def _calc_prob_death(spp, d):
    #get the fitness values (while also setting all individ.fit attributes)
    w = spp._calc_fitness()
    death_probs = 1-(1-d)*w
    assert (death_probs >= 0).all() and (death_probs <= 1).all(), ("Some "
                        "death-probability values outside the 0-to-1 range.")
    return(death_probs)




#!/usr/bin/python
#individual.py

# flake8: noqa

'''
Defines the Individual class, with its associated methods and supporting
functions
'''

#geonomics imports
from geonomics.ops.movement import _do_dispersal
from geonomics.ops.selection import _calc_phenotype

#other imports
import numpy as np
import numpy.random as r


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

class Individual:
    """
    Representation of an individual of a given species.

    Multiple Individuals are collected as serial integer-keyed values within a
    Species dict. Those serial indices continually increment through model time
    as new Individuals (i.e. offspring) are produced, such that no two
    Individuals within the full history of a simulation will ever have the same
    index number.

    Attributes
    ----------

        NOTE: For more detail, see the documentation for the parameters
              that correspond to many of the following attributes.

        age:
            The Individual's age (expressed in timesteps since its birth
            timestep, when the individual was instantiated at age 0)

        e:
            The Individual's current environmental values, organized as a
            1d numpy array of length == len(Landscape), where the value stored
            at Individual.e[i] gives the Individual's current environmental
            value for Layer number i (i.e. Landscape[i])

        fit:
            The Individual's current fitness (calculated as a combination of
            the Individual's phenotypes for all traits, its current
            environmental values for all corresponding Landscape Layers, and
            any deleterious loci for which it has '1' alleles; see
            documentation for further details)

        g:
            The Individual's genotype array,
            stored as an L_g x 2 numpy array, where L_g is the number
            of loci being tracked by each Individual in the Species.

            If tskit is being used then Individuals only carry
            copies of their non-neutral genotypes
            (stored in this attribute); this is a computational optimization,
            as it allows fitness-based operations to be calculated
            quickly on the fly (using this attribute's data), while minimizing
            the memory required to stored the full (neutral and non-netural)
            genotypes of all current Individuals and their ancestors (i.e. the
            'spatial pedigree' stored in the succinct set of tskit tables).
            The successive rows in this array store the genotypes corresponding
            to the loci indicated by the successive locus numbers in
            Species.GenomicArchitecture.nonneut_loci.

            If tskit is not being used then this array will store genotypes
            for all loci in the simulated genome (i.e. it will be of shape
            L x 2, where L is genome-length parameter in the this Individual's
            Species' section of the paramters file).

        sex:
            The Individual's sex (0=female, 1=male; None, if the Species is
            unsexed)
        x:
            The Individual's current x (i.e. longitudinal) position
            (in continuous space, bounded within [0, Landscape.dims[0]])

        y:
            The Individual's current y (i.e. latitudinal) position
            (in continuous space, bounded within [0, Landscape.dims[1]])

        z:
            The Individual's phenotypes for each, organized as a
            1d numpy array of length == len(Species.traits), where the value
            stored at Individual.z[i] gives the Individual's fitness for
            Trait number i (i.e. Species.traits[i]).


    """
    def __init__(self, idx, x, y, age=0, new_genome=None, sex=None):
        self.idx = idx
        #individual's x-ploid genome (NOTE: make np.int8 to minimize mem)
        if new_genome is not None:
            self.g = np.int8(new_genome)
        else:
            self.g = new_genome
        #x and y coords
        self.x = float(x)
        self.y = float(y)
        #set individual's sex to that supplied, if supplied
        if sex:
            self.sex = sex
        #else, set it to a bernoulli r.v., p = 0.5;  0 --> female)
        else:
            self.sex = r.binomial(1, 0.5)
        self.age = age
        self.e = None
        self.z = []
        self.fit = None

        # add attributes to hold the Individual's tskit Individuals id
        # and Nodes ids
        self._individuals_tab_id = None
        self._nodes_tab_ids = {}

        assert type(self.x) == float and self.x >= 0, ("invalid value "
                                "for x: %s, %s") % (str(self.x), type(self.x))
        assert type(self.y) == float and self.y >= 0, ("invalid value "
                                "for y: %s, %s") % (str(self.y), type(self.y))
        assert self.sex == None or self.sex in [0,1]
        assert type(self.age) == int


    #####################
    ### OTHER METHODS ###
    #####################

    # function to increment age by one
    def _set_age_stage(self):
        self.age += 1

    # set the individual's position
    def _set_pos(self, x_pos, y_pos):
        self.x = x_pos
        self.y = y_pos

    # set the individual's current environmental values (attribute e)
    def _set_e(self, e):
        self.e = e

    # set the individual's phenotype (attribute z) for all traits
    def _set_z(self, genomic_architecture):
        self.z = [_calc_phenotype(self, genomic_architecture,
            trait_num) for trait_num in genomic_architecture.traits]

    # set the individual's fitness
    def _set_fit(self, fit):
        self.fit = fit

    # set the individual's genome
    def _set_g(self, genome):
        self.g = genome

    # add a row of zeros to the individual's genome at the given index
    def _add_new_locus(self, idx):
        self.g = np.vstack((self.g[:idx, :],
                            [0, 0],
                            self.g[idx:, :]))

    # set the individual's tskit.TableCollection.nodes table's node ids
    # NOTE: node ids must be fed in 0-to-x order, where x is the species'
    # ploidy, because they will be assigned to the 0, 1, ..., x-keyed values
    # of the Individual._nodes_tab_ids dict according to that order
    def _set_nodes_tab_ids(self, *node_ids):
        self._nodes_tab_ids.update({i:id for i, id in enumerate(node_ids)})

    # get the individual's x,y location as a tuple
    def _get_loc(self):
        return(self.x, self.y)


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

def _make_individual(idx, offspring=True, dim=None, genomic_architecture=None,
        new_genome=None, sex=None, parental_centerpoint=None,
        age=0, burn=False):

    """Create a new individual.

        If it is to have a genome, that can be created from either
        an instance of genome.Genomic_Architecture (i.e. for
        a newly simulated individual) or both the genomic architecture
        and a numpy.ndarray genome (e.g. for a new offspring).

        It will be assigned x- and y-coordinates, using the
        dim argument (i.e. the landscape dimensionality)
        if it is a newly simulated individual rather than offspring.

        It will be assigned a random sex, unless sex is provided.

        It will be assigned age 0, unless specifically fed otherwise.
        """
    #set the x,y location of the individual
    if offspring:
        #get a starting position from the parental_centerpoint
        x,y = _do_dispersal(parental_centerpoint)
    else:
        #randomly assign individual a valid starting location
        x,y = r.rand(2)*dim
        #clip to 0.01 under the dimensions, so that for landscapes even up to
        #~10000 on a side (this is bigger than I expect most users would
        #want to run) np.float32 can't return a number rounded up to
        #the dimension itself if an extremely high value is drawn
        x = np.clip(x, 0, dim[0]-0.001)
        y = np.clip(y, 0, dim[1]-0.001)

    #set the sex, if necessary
    if sex is None:
        #NOTE: For now sex randomly chosen at 50/50. Change if
        #decide to implement sex chroms, or spp.sex_ratio
        sex = r.binomial(1,0.5)

    individual = Individual(idx=idx, x=x, y=y, age=age,
                            new_genome=new_genome, sex=sex)
    return individual

